from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mixemy.repositories import BaseRepository
    from mixemy.services import BaseService
    from mixemy.types import (
        BaseModelType,
        CreateSchemaType,
        FilterSchemaType,
        OutputSchemaType,
        UpdateSchemaType,
    )


class MixemyError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class MixemySetupError(MixemyError):
    def __init__(
        self,
        component: object,
        component_name: str,
        undefined_field: str,
        message: str | None = None,
    ) -> None:
        if message is None:
            message = f"{component_name.capitalize()} {component} has undefined field '{undefined_field}'.\nThis probably needs to be defined as a class attribute."
        self.component = component
        self.component_name = component_name
        self.undefined_field = undefined_field
        super().__init__(message=message)


class MixemyRepositorySetupError(MixemySetupError):
    def __init__(
        self,
        repository: "BaseRepository[BaseModelType]",
        undefined_field: str,
        message: str | None = None,
    ) -> None:
        self.repository = repository
        super().__init__(
            message=message,
            component=repository,
            undefined_field=undefined_field,
            component_name="repository",
        )


class MixemyServiceSetupError(MixemySetupError):
    def __init__(
        self,
        service: "BaseService[BaseModelType, CreateSchemaType, UpdateSchemaType, FilterSchemaType, OutputSchemaType]",
        undefined_field: str,
        message: str | None = None,
    ) -> None:
        self.service = service
        super().__init__(
            message=message,
            component=service,
            undefined_field=undefined_field,
            component_name="service",
        )



from abc import ABC

from mixemy.repositories.asyncio import IdAsyncRepository
from mixemy.services.asyncio._base import BaseAsyncService
from mixemy.types import (
    ID,
    CreateSchemaType,
    FilterSchemaType,
    IdModelType,
    OutputSchemaType,
    UpdateSchemaType,
)


class IdAsyncService(
    BaseAsyncService[
        IdModelType,
        CreateSchemaType,
        UpdateSchemaType,
        FilterSchemaType,
        OutputSchemaType,
    ],
    ABC,
):
    repository_type: type[IdAsyncRepository[IdModelType]]  # pyright: ignore[reportIncompatibleVariableOverride] - https://github.com/python/typing/issues/548
    repository: IdAsyncRepository[IdModelType]

    async def read(self, id: ID) -> OutputSchemaType | None:
        return (
            self._to_schema(model=model)
            if (
                model := await self.repository.read_by_id(
                    db_session=self.db_session, id=id, raise_on_empty=False
                )
            )
            else None
        )

    async def update(self, id: ID, object_in: UpdateSchemaType) -> OutputSchemaType:
        return self._to_schema(
            model=await self.repository.update_by_id(
                db_session=self.db_session,
                id=id,
                object_in=object_in,
            )
        )

    async def delete(self, id: ID) -> None:
        await self.repository.delete_by_id(
            db_session=self.db_session, id=id, raise_on_empty=False
        )



from asyncio import current_task
from collections.abc import AsyncGenerator, Generator

import pytest
import pytest_asyncio
from sqlalchemy import Engine, String, create_engine, make_url
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_scoped_session,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.orm import Mapped, Session, mapped_column, scoped_session, sessionmaker
from testcontainers.postgres import PostgresContainer


@pytest.fixture(scope="module")
def db_container() -> Generator[PostgresContainer, None, None]:
    with PostgresContainer("postgres:latest") as postgres:
        yield postgres


@pytest.fixture(scope="module")
def engine(db_container: PostgresContainer) -> Engine:
    return create_engine(make_url(db_container.get_connection_url()))


@pytest.fixture(scope="module")
def scoped_session_maker(engine: Engine) -> scoped_session[Session]:
    return scoped_session(sessionmaker(bind=engine))


@pytest.fixture(scope="module")
def async_engine(db_container: PostgresContainer) -> AsyncEngine:
    return create_async_engine(
        make_url(db_container.get_connection_url(driver="asyncpg"))
    )


@pytest.fixture(scope="module")
def async_scoped_session_maker(
    async_engine: AsyncEngine,
) -> async_scoped_session[AsyncSession]:
    return async_scoped_session(
        async_sessionmaker(bind=async_engine, expire_on_commit=False),
        scopefunc=current_task,
    )


@pytest.fixture
def session(
    scoped_session_maker: scoped_session[Session],
) -> Generator[Session]:
    session = scoped_session_maker()
    try:
        yield session
    finally:
        session.close()


@pytest_asyncio.fixture(scope="function")  # pyright: ignore[reportUntypedFunctionDecorator,reportUnknownMemberType]
async def async_session(
    async_scoped_session_maker: async_scoped_session[AsyncSession],
) -> AsyncGenerator[AsyncSession]:
    session = async_scoped_session_maker()
    try:
        yield session
    finally:
        await session.close()


@pytest.fixture(scope="module")
def init_db(engine: Engine) -> None:
    from mixemy import models

    class AsyncItemModel(models.IdAuditModel):
        __table_args__ = {"extend_existing": True}  # noqa: RUF012
        value: Mapped[str] = mapped_column(String)

    class ItemModel(models.IdAuditModel):
        __table_args__ = {"extend_existing": True}  # noqa: RUF012
        value: Mapped[str] = mapped_column(String)
        nullable_value: Mapped[str | None] = mapped_column(String, nullable=True)

    ItemModel.__table__.create(bind=engine)  # type: ignore - Only for testing
    AsyncItemModel.__table__.create(bind=engine)  # type: ignore - Only for testing



from typing import Any, TypeVar

from mixemy.models import BaseModel
from mixemy.schemas import BaseSchema

TO_MODEL = TypeVar("TO_MODEL", bound=BaseModel)
TO_SCHEMA = TypeVar("TO_SCHEMA", bound=BaseSchema)


def unpack_schema(
    schema: BaseSchema,
    exclude_unset: bool = True,
    exclude: set[str] | None = None,
) -> dict[str, Any]:
    return schema.model_dump(exclude_unset=exclude_unset, exclude=exclude)


def to_model(model: type[TO_MODEL], schema: BaseSchema) -> TO_MODEL:
    return model(**unpack_schema(schema=schema))


def to_schema(model: BaseModel, schema: type[TO_SCHEMA]) -> TO_SCHEMA:
    return schema.model_validate(model)



from . import asyncio, sync
from ._base import BaseRepository
from .asyncio import (
    AuditAsyncRepository,
    BaseAsyncRepository,
    IdAsyncRepository,
    IdAuditAsyncRepository,
)
from .sync import (
    AuditSyncRepository,
    BaseSyncRepository,
    IdAuditSyncRepository,
    IdSyncRepository,
)

__all__ = [
    "AuditAsyncRepository",
    "AuditSyncRepository",
    "BaseAsyncRepository",
    "BaseRepository",
    "BaseSyncRepository",
    "IdAsyncRepository",
    "IdAuditAsyncRepository",
    "IdAuditSyncRepository",
    "IdSyncRepository",
    "asyncio",
    "sync",
]



from abc import ABC
from typing import Literal, overload

from sqlalchemy import delete
from sqlalchemy.orm import Session

from mixemy.repositories._base import IdRepository
from mixemy.repositories.sync._base import BaseSyncRepository
from mixemy.schemas import InputSchema
from mixemy.types import ID, IdModelType


class IdSyncRepository(BaseSyncRepository[IdModelType], IdRepository[IdModelType], ABC):
    def update_by_id(
        self,
        db_session: Session,
        id: ID,
        object_in: InputSchema,
    ) -> IdModelType:
        return self.update(
            db_session=db_session,
            db_object=self.read_by_id(
                db_session=db_session, id=id, raise_on_empty=True
            ),
            object_in=object_in,
        )

    @overload
    def read_by_id(
        self, db_session: Session, id: ID, raise_on_empty: Literal[True]
    ) -> IdModelType: ...

    @overload
    def read_by_id(
        self, db_session: Session, id: ID, raise_on_empty: Literal[False]
    ) -> IdModelType | None: ...

    @overload
    def read_by_id(
        self, db_session: Session, id: ID, raise_on_empty: bool
    ) -> IdModelType | None: ...

    @overload
    def read_by_id(
        self, db_session: Session, id: ID, raise_on_empty: bool = False
    ) -> IdModelType | None: ...

    def read_by_id(
        self, db_session: Session, id: ID, raise_on_empty: bool = False
    ) -> IdModelType | None:
        db_object = db_session.get(self.model, id)

        if db_object is None and raise_on_empty:
            msg = f"Could not find {self.model} of id {id}"
            raise ValueError(msg)

        return db_object

    def delete_by_id(
        self,
        db_session: Session,
        id: ID,
        raise_on_empty: bool,
    ) -> None:
        if raise_on_empty:
            self._delete_refreshing(
                db_session=db_session,
                db_object=self.read_by_id(
                    db_session=db_session,
                    id=id,
                    raise_on_empty=raise_on_empty,
                ),
            )
        else:
            statement = delete(self.model).where(self.model.id == id)
            db_session.execute(statement)
            db_session.commit()



from typing import overload
from uuid import UUID


@overload
def serialize_uuid(data: UUID) -> str: ...


@overload
def serialize_uuid(data: None) -> None: ...


def serialize_uuid(data: UUID | None) -> str | None:
    if data is None:
        return data

    return str(data)



from abc import ABC

from mixemy.repositories._base import AuditRepository
from mixemy.repositories.sync._base import BaseSyncRepository
from mixemy.types import AuditModelType


class AuditSyncRepository(
    BaseSyncRepository[AuditModelType], AuditRepository[AuditModelType], ABC
):
    pass



from . import asyncio, sync
from ._base import BaseService
from .asyncio import (
    AuditAsyncService,
    BaseAsyncService,
    IdAsyncService,
    IdAuditAsyncService,
)
from .sync import AuditSyncService, BaseSyncService, IdAuditSyncService, IdSyncService

__all__ = [
    "AuditAsyncService",
    "AuditSyncService",
    "BaseAsyncService",
    "BaseService",
    "BaseSyncService",
    "IdAsyncService",
    "IdAuditAsyncService",
    "IdAuditSyncService",
    "IdSyncService",
    "asyncio",
    "sync",
]



from ._audit import AuditAsyncService
from ._base import BaseAsyncService
from ._id import IdAsyncService
from ._id_audit import IdAuditAsyncService

__all__ = [
    "AuditAsyncService",
    "BaseAsyncService",
    "IdAsyncService",
    "IdAuditAsyncService",
]



from abc import ABC
from typing import Any, Generic

from sqlalchemy import Select, select

from mixemy._exceptions import MixemyRepositorySetupError
from mixemy.schemas import InputSchema
from mixemy.schemas.paginations import PaginationFields, PaginationFilter
from mixemy.types import BaseModelType
from mixemy.utils import unpack_schema


class BaseRepository(Generic[BaseModelType], ABC):
    model_type: type[BaseModelType]

    def __init__(self) -> None:
        self._verify_init()
        self.model = self.model_type

    def _add_after_filter(
        self, statement: Select[Any], filters: InputSchema | None
    ) -> Select[Any]:
        if isinstance(filters, PaginationFilter):
            statement = statement.offset(filters.offset).limit(filters.limit)

        return statement

    def _add_before_filter(
        self, statement: Select[Any], filters: InputSchema | None
    ) -> Select[Any]:
        if filters is not None:
            for item, value in unpack_schema(
                schema=filters, exclude=PaginationFields
            ).items():
                if hasattr(self.model, item):
                    if isinstance(value, list):
                        statement = statement.where(
                            getattr(self.model, item).in_(value)
                        )
                    else:
                        statement = statement.where(getattr(self.model, item) == value)

        return statement

    def _add_filters(
        self, statement: Select[Any] | None, filters: InputSchema | None
    ) -> Select[Any]:
        if statement is None:
            statement = select(self.model)
        return self._add_after_filter(
            statement=self._add_before_filter(statement=statement, filters=filters),
            filters=filters,
        )

    @staticmethod
    def _update_db_object(db_object: BaseModelType, object_in: InputSchema) -> None:
        for field, value in unpack_schema(schema=object_in).items():
            if hasattr(db_object, field):
                setattr(db_object, field, value)

    def _verify_init(self) -> None:
        for field in ["model_type"]:
            if not hasattr(self, field):
                raise MixemyRepositorySetupError(repository=self, undefined_field=field)



from abc import ABC
from typing import Any, override

from sqlalchemy import Select, asc, desc

from mixemy.repositories._base._base import BaseRepository
from mixemy.schemas import InputSchema
from mixemy.schemas.paginations import AuditPaginationFilter, OrderDirection
from mixemy.types import (
    AuditModelType,
)


class AuditRepository(BaseRepository[AuditModelType], ABC):
    @override
    def _add_after_filter(
        self, statement: Select[Any], filters: InputSchema | None
    ) -> Select[Any]:
        statement = super()._add_after_filter(statement=statement, filters=filters)
        if isinstance(filters, AuditPaginationFilter):
            match filters.order_direction:
                case OrderDirection.ASC:
                    statement = statement.order_by(asc(filters.order_by))
                case OrderDirection.DESC:
                    statement = statement.order_by(desc(filters.order_by))

        return statement



from ._audit import AuditSyncService
from ._base import BaseSyncService
from ._id import IdSyncService
from ._id_audit import IdAuditSyncService

__all__ = [
    "AuditSyncService",
    "BaseSyncService",
    "IdAuditSyncService",
    "IdSyncService",
]



from abc import ABC

from mixemy.repositories.sync import AuditSyncRepository
from mixemy.services.sync._base import BaseSyncService
from mixemy.types import (
    AuditModelType,
    CreateSchemaType,
    FilterSchemaType,
    OutputSchemaType,
    UpdateSchemaType,
)


class AuditSyncService(
    BaseSyncService[
        AuditModelType,
        CreateSchemaType,
        UpdateSchemaType,
        FilterSchemaType,
        OutputSchemaType,
    ],
    ABC,
):
    repository_type: type[AuditSyncRepository[AuditModelType]]  # pyright: ignore[reportIncompatibleVariableOverride] - https://github.com/python/typing/issues/548
    repository: AuditSyncRepository[AuditModelType]



from abc import ABC

from sqlalchemy.ext.asyncio import AsyncSession

from mixemy.repositories.asyncio import BaseAsyncRepository
from mixemy.services._base import BaseService
from mixemy.types import (
    BaseModelType,
    CreateSchemaType,
    FilterSchemaType,
    OutputSchemaType,
    UpdateSchemaType,
)


class BaseAsyncService(
    BaseService[
        BaseModelType,
        CreateSchemaType,
        UpdateSchemaType,
        FilterSchemaType,
        OutputSchemaType,
    ],
    ABC,
):
    repository_type: type[BaseAsyncRepository[BaseModelType]]

    def __init__(self, db_session: AsyncSession) -> None:
        super().__init__()
        self.repository = self.repository_type()
        self.model = self.repository.model
        self.db_session = db_session

    async def create(self, object_in: CreateSchemaType) -> OutputSchemaType:
        return self._to_schema(
            model=await self.repository.create(
                db_session=self.db_session, db_object=self._to_model(schema=object_in)
            )
        )

    async def read_multi(
        self,
        filters: FilterSchemaType | None = None,
    ) -> list[OutputSchemaType]:
        return [
            self._to_schema(model=model)
            for model in await self.repository.read_multi(
                db_session=self.db_session,
                filters=filters,
            )
        ]



from ._audit import AuditPaginationFilter
from ._base import PaginationFilter
from ._order_enums import OrderBy, OrderDirection

PaginationFields = {"limit", "offset", "order_by", "order_direction"}

__all__ = [
    "AuditPaginationFilter",
    "OrderBy",
    "OrderDirection",
    "PaginationFields",
    "PaginationFilter",
]



from mixemy.models._audit import AuditModel
from mixemy.models._id import IdModel


class IdAuditModel(IdModel, AuditModel):
    __abstract__ = True

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} created_at={self.created_at} updated_at={self.updated_at}>"



from typing import TYPE_CHECKING, TypeVar

# from mixemy.schemas import InputSchema
# from mixemy.schemas.paginations import AuditPaginationFilter, PaginationFilter

if TYPE_CHECKING:
    from mixemy.schemas import InputSchema, OutputSchema
    from mixemy.schemas.paginations import AuditPaginationFilter, PaginationFilter

CreateSchemaType = TypeVar("CreateSchemaType", bound="InputSchema")
UpdateSchemaType = TypeVar("UpdateSchemaType", bound="InputSchema")
OutputSchemaType = TypeVar("OutputSchemaType", bound="OutputSchema")

FilterSchemaType = TypeVar("FilterSchemaType", bound="InputSchema")

# PaginationSchemaType = TypeVar("PaginationSchemaType", bound=PaginationFilter)
# type FilterSchemaType = "InputSchema"
type PaginationSchemaType = "PaginationFilter"
type AuditPaginationSchemaType = "AuditPaginationFilter"



import pytest
from sqlalchemy import String
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Mapped, mapped_column


@pytest.mark.database
@pytest.mark.integration
@pytest.mark.asyncio
async def test_main(async_session: AsyncSession, init_db: None) -> None:
    from mixemy import models, repositories, schemas, services

    class AsyncItemModel(models.IdAuditModel):
        __table_args__ = {"extend_existing": True}  # noqa: RUF012
        value: Mapped[str] = mapped_column(String)

    class ItemInput(schemas.InputSchema):
        value: str

    class ItemOutput(schemas.IdAuditOutputSchema):
        value: str

    class ItemRepository(repositories.IdAuditAsyncRepository[AsyncItemModel]):
        model_type = AsyncItemModel

    class ItemService(
        services.IdAuditAsyncService[
            AsyncItemModel, ItemInput, ItemInput, ItemInput, ItemOutput
        ]
    ):
        repository_type = ItemRepository
        output_schema_type = ItemOutput

    item_service = ItemService(db_session=async_session)

    test_one = ItemInput(value="test_one")
    test_two = ItemInput(value="test_two")

    item_one = await item_service.create(object_in=test_one)
    item_two = await item_service.create(object_in=test_two)

    assert item_one.value == "test_one"
    assert item_two.value == "test_two"

    item_one = await item_service.read(id=item_one.id)
    item_two = await item_service.read(id=item_two.id)

    assert item_one is not None
    assert item_two is not None
    assert item_one.value == "test_one"
    assert item_two.value == "test_two"

    item_one = await item_service.update(id=item_one.id, object_in=test_two)

    assert item_one.value == "test_two"

    await item_service.delete(id=item_one.id)

    item_one = await item_service.read(id=item_one.id)
    item_two = await item_service.read(id=item_two.id)

    assert item_one is None
    assert item_two is not None
    assert item_two.value == "test_two"

    await item_service.delete(id=item_two.id)

    item_two = await item_service.read(id=item_two.id)

    assert item_two is None



from . import paginations, serializers
from ._audit_output import AuditOutputSchema
from ._base import BaseSchema
from ._id_audit_output import IdAuditOutputSchema
from ._id_output import IdOutputSchema
from ._input import InputSchema
from ._output import OutputSchema

__all__ = [
    "AuditOutputSchema",
    "BaseSchema",
    "IdAuditOutputSchema",
    "IdOutputSchema",
    "InputSchema",
    "OutputSchema",
    "paginations",
    "serializers",
]
